/*
 * Copyright (c) 1983 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static char sccsid[] = "@(#)announce.c	5.6 (Berkeley) 6/18/88";
#endif /* not lint */

#include <sys/types.h>
#include <sys/stat.h>
#include <sgtty.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <stdio.h>
#include <sys/wait.h>
#include <errno.h>
#include <syslog.h>
#include <utmp.h>
#include "talkd.h"

#ifndef TIOCNOTTY
#define	TIOCNOTTY	_IO('t', 113)		/* void tty association */
#endif

#define MAX_WINDOWS    10

extern	int errno;
extern	char hostname[];

/*
 * Announce an invitation to talk.
 *
 * Because the tty driver insists on attaching a terminal-less
 * process to any terminal that it writes on, we must fork a child
 * to protect ourselves
 */
announce(request, remote_machine)
	CTL_MSG *request;
	char *remote_machine;
{
	int pid, val, status;

	if (pid = fork()) {
		/* we are the parent, so wait for the child */
		if (pid == -1)		/* the fork failed */
			return (FAILED);
		do {
			val = wait(&status);
			if (val == -1) {
				if (errno == EINTR)
					continue;
				/* shouldn't happen */
				syslog(LOG_WARNING, "announce: wait: %m");
				return (FAILED);
			}
		} while (val != pid);
		if (status&0377 > 0)	/* we were killed by some signal */
			return (FAILED);
		/* Get the second byte, this is the exit/return code */
		return ((status >> 8) & 0377);
	}
	/* we are the child, go and do it */
	_exit(announce_proc(request, remote_machine));
}
	
/*
 * See if the user is accepting messages. If so, announce that 
 * a talk is requested.
 */
announce_proc(request, remote_machine)
	CTL_MSG *request;
	char *remote_machine;
{
	int pid, status;
	struct stat stbuf;
#ifdef __hpux
        int term_n=0, access_ok[MAX_WINDOWS], 
            permission[MAX_WINDOWS], cont=0, acc_den = 0;
	int i, j;
	char full_tty[MAX_WINDOWS][32];
	FILE *tf[MAX_WINDOWS];
	struct utmp *t;

	while ( (t=getutent())!=NULL ) {
		if ( t->ut_type==USER_PROCESS && strncmp(t->ut_user,request->r_name,sizeof(t->ut_user))==0 ) {
				
			sprintf(full_tty[term_n],"/dev/%s",t->ut_line);
			access_ok[term_n] = 
				(access(full_tty[term_n], 0) == 0) ;
			permission[term_n] = 
				!((tf[term_n] = fopen(full_tty[term_n], "w")) == NULL);
			setsid();
			/* ioctl(fileno(tf[term_n]), TIOCNOTTY, (struct sgttyb *) 0); */
	/*
	 * On first tty open, the server will have
	 * it's pgrp set, so disconnect us from the
	 * tty before we catch a signal.
	 */
			permission[term_n] = 
				!(fstat(fileno(tf[term_n]), &stbuf) < 0);
			permission[term_n] = 
				!((stbuf.st_mode&020) == 0);
			term_n++;
#ifdef NO_CONSOLE
			if ( strcmp(t->ut_line,"console")==0 ) 
				term_n--;
#endif
			if (term_n == MAX_WINDOWS) break;
		}
	}
	for (i=0;i<term_n;i++) {
		cont = ((access_ok[i])&&(permission[i]));
		if (!cont) {
			acc_den = (!access_ok[i]);
			for (j=i;j<term_n-1;j++) {
				access_ok[j] = access_ok[j+1];
				permission[j] = permission[j+1];
				tf[j] = tf[j+1];
			}
			term_n--;
		}
	}
	cont = (term_n>0);
	if (!cont) {
		if (acc_den) return (FAILED);
		else return (PERMISSION_DENIED);
	}
	for (i=0; i<term_n;i++) {
		if ( (access_ok[i])&&(permission[i]) ) {
			print_mesg(tf[i], request, remote_machine);
			fclose(tf[i]);
		}
	}
#else /* __hpux */
        char full_tty[32];
        FILE *tf;

        (void)sprintf(full_tty, "/dev/%s", request->r_tty);
        if (access(full_tty, 0) != 0)
                return (FAILED);
        if ((tf = fopen(full_tty, "w")) == NULL)
                return (PERMISSION_DENIED);
        /*
         * On first tty open, the server will have
         * it's pgrp set, so disconnect us from the
         * tty before we catch a signal.
         */
        ioctl(fileno(tf), TIOCNOTTY, (struct sgttyb *) 0);
        if (fstat(fileno(tf), &stbuf) < 0)
                return (PERMISSION_DENIED);
        if ((stbuf.st_mode&020) == 0)
                return (PERMISSION_DENIED);
        print_mesg(tf, request, remote_machine);
        fclose(tf);
#endif /* __hpux */

	return (SUCCESS);
}

#define max(a,b) ( (a) > (b) ? (a) : (b) )
#define N_LINES 5
#define N_CHARS 120

/*
 * Build a block of characters containing the message. 
 * It is sent blank filled and in a single block to
 * try to keep the message in one piece if the recipient
 * in in vi at the time
 */
print_mesg(tf, request, remote_machine)
	FILE *tf;
	CTL_MSG *request;
	char *remote_machine;
{
	struct timeval clock;
	struct timezone zone;
	struct tm *localtime();
	struct tm *localclock;
	char line_buf[N_LINES][N_CHARS];
	int sizes[N_LINES];
	char big_buf[N_LINES*N_CHARS];
	char *bptr, *lptr;
	int i, j, max_size;

	i = 0;
	max_size = 0;
	gettimeofday(&clock, &zone);
#ifdef __hpux
	localclock = localtime( (const time_t *)&clock.tv_sec );
#else
	localclock = localtime( &clock.tv_sec );
#endif
	(void)sprintf(line_buf[i], " ");
	sizes[i] = strlen(line_buf[i]);
	max_size = max(max_size, sizes[i]);
	i++;
	(void)sprintf(line_buf[i],"Message from nTalk_Daemon@%s at %d:%02d ...",
	hostname, localclock->tm_hour , localclock->tm_min );
	sizes[i] = strlen(line_buf[i]);
	max_size = max(max_size, sizes[i]);
	i++;
	(void)sprintf(line_buf[i], "ntalkd: connection requested by %s@%s.",
		request->l_name, remote_machine);
	sizes[i] = strlen(line_buf[i]);
	max_size = max(max_size, sizes[i]);
	i++;
	(void)sprintf(line_buf[i], "ntalkd: respond with:  ytalk %s@%s",
		request->l_name, remote_machine);
	sizes[i] = strlen(line_buf[i]);
	max_size = max(max_size, sizes[i]);
	i++;
	(void)sprintf(line_buf[i], " ");
	sizes[i] = strlen(line_buf[i]);
	max_size = max(max_size, sizes[i]);
	i++;
	bptr = big_buf;
	*bptr++ = ''; /* send something to wake them up */
	*bptr++ = '\r';	/* add a \r in case of raw mode */
	*bptr++ = '\n';
	for (i = 0; i < N_LINES; i++) {
		/* copy the line into the big buffer */
		lptr = line_buf[i];
		while (*lptr != '\0')
			*(bptr++) = *(lptr++);
		/* pad out the rest of the lines with blanks */
		for (j = sizes[i]; j < max_size + 2; j++)
			*(bptr++) = ' ';
		*(bptr++) = '\r';	/* add a \r in case of raw mode */
		*(bptr++) = '\n';
	}
	*bptr = '\0';
	fprintf(tf, big_buf);
	fflush(tf);
#ifdef __hpux
        setsid();
#else
	ioctl(fileno(tf), TIOCNOTTY, (struct sgttyb *) 0);
#endif
}
